#include <gtk/gtk.h>
#include <stdio.h>
#include <string.h>

/*Подключение заголовочного файла с данными ресурсов*/
#include "gwhois-resources.h"

#define APPLICATION_WHOIS_PORT           43
#define APPLICATION_SOCKET_TIMEOUT        5
#define APPLICATION_SERVER_LIST_RESOURCE "/org/gtk/GWhois/whois-server-list.xml"
#define APPLICATION_CSS_RESOURCE         "/org/gtk/GWhois/gwhois.css"

/*Перечисление идентификаторов элементов файла со списком серверов формата XML*/
enum {
	APPLICATION_XML_PARAMETER_DOMAIN = 0,
	APPLICATION_XML_PARAMETER_WHOISSERVER,
	APPLICATION_XML_PARAMETER_STATE,
	APPLICATION_XML_PARAMETER_UNKNOWN
};

/*Идентификаторы классов сообщений*/
enum {
	APPLICATION_PROCESS_START = 0,
	APPLICATION_PROCESS_END,
	APPLICATION_PROCESS_ERROR
};

/*Данные приложения*/
typedef struct {
	/*Объект приложения*/
	GtkApplication *app;
	/*Хэш-таблица с именами серверов whois*/
	GHashTable *whois_servers;
	/*Текущий уровень вложенности описаний доменов*/
	guint current_whois_level;
	/*Текущее имя сервера whois*/
	gchar *current_whois_name;
	/*Текущий сервер whois*/
	gchar *current_whois_server;
	/*Идентификатор текущего параметра файла со списком серверов формата XML*/
	guint xmlparameter;
	/*Виджет основного окна приложения*/
	GtkWidget *mainwin; 
	/*Виджет для ввода имени домена*/
	GtkWidget *domain_entry;
	/*Идентификатор обработчика сигнала изменения содержимого поля ввода имени домена*/
	gulong domain_entry_activate_signal;
	/*Состояние обработчика сигнала активации поля ввода имени домена*/
	gboolean domain_entry_activation;
	/*Виджет кнопки для получения информации*/
	GtkWidget *search_button;
	/*Виджет поля для редактирования текста*/
	GtkWidget *whois_textview;
	/*Виджет строки состояния*/
	GtkWidget *statusbar;
	/*Идентификатор последнего сообщения виджета строки состояния*/
	guint statusmsgid;
	/*Идентификатор контекста виджета строки состояния*/
	guint statusconid;
	/*Виджет индикатора выполнения задачи*/
	GtkWidget *spinner;
} appdata;

/*Функция извлечения имени доменной зоны из строки полного имени домена*/
static gchar *application_get_domain_name(gchar *fulldomain)
{
	gchar *domain;
	
	if (fulldomain != NULL) {
		/*Поиск символа точки в полном имени домена начиная с конца строки*/
		domain = strrchr(fulldomain, '.');
		if (domain != NULL) {
			/*Возврат имени доменной зоны, начинающегося с первого символа
			 *после найденного символа точки*/
			if ((domain[0] != '\0') && (domain[1] != '\0')) {
				return domain + 1;
			}
		}
	}
	
	return NULL;
}

/*Функция освобождения памяти, зарезервированной для хранения имени сервера whois*/
static void application_free_server_name(gpointer data)
{
	g_free(data);
}

/*Функция освобождения памяти, зарезервированной для хранения имени доменной зоны,
 *информация о которой предоставляется сервером whois*/
static void application_free_domain_name(gpointer data)
{
	g_free(data);
}

/*Функция начала разбора элемента списка серверов whois*/
static void application_xml_get_start_element(GMarkupParseContext *context, const gchar *element_name, const gchar **attribute_names, const gchar **attribute_values, gpointer user_data, GError **error)
{
	appdata *data;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Установка соответствующего идентификатора элемента для каждого из имен элементов*/
	if (g_str_equal(element_name, "domainList")) {
		data->xmlparameter = APPLICATION_XML_PARAMETER_UNKNOWN;
	} else if (g_str_equal(element_name, "domain")) {
		/*Сохранение имени доменной зоны*/
		if (data->current_whois_name != NULL) {
			g_free(data->current_whois_name);
		}
		data->current_whois_name = g_strdup(attribute_values[0]);
		data->xmlparameter = APPLICATION_XML_PARAMETER_DOMAIN;
	} else if (g_str_equal(element_name, "whoisServer")) {
		/*Сохранение имени сервера whois*/
		if (data->current_whois_server != NULL) {
			g_free(data->current_whois_server);
		}
		data->current_whois_server = g_strdup(attribute_values[0]);
		data->xmlparameter = APPLICATION_XML_PARAMETER_WHOISSERVER;
	}else {
		/*Неизвестный элемент*/
		data->xmlparameter = APPLICATION_XML_PARAMETER_UNKNOWN;
	}
}

/*Функция разбора текста списка серверов whois*/
static void application_xml_get_value(GMarkupParseContext *context, const gchar *text, gsize text_len, gpointer user_data, GError **error)
{
	appdata *data;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Идентификатор неизвестного элемента*/
	if (data->xmlparameter == APPLICATION_XML_PARAMETER_UNKNOWN) return;
	
	switch (data->xmlparameter) {
		case APPLICATION_XML_PARAMETER_DOMAIN:
			/*Увеличение текущего уровня вложенности*/
			data->current_whois_level++;
			break;
		case APPLICATION_XML_PARAMETER_WHOISSERVER:
			if ((data->current_whois_name != NULL) && (data->current_whois_server != NULL)) {
				/*Добавление информации о серере whois для доменной зоны в хэш-таблицу при
				  условии получения всей информации (текущий уровень вложенности равен 1)*/
				if (data->current_whois_level == 1) {
					g_hash_table_insert(data->whois_servers, g_strdup(data->current_whois_name), g_strdup(data->current_whois_server));
				}
			}
			break;
		case APPLICATION_XML_PARAMETER_STATE:
			
			break;
		default:
			break;
	}
}

/*Функция окончания разбора элемента списка серверов whois*/
static void application_xml_get_end_element(GMarkupParseContext *context, const gchar *element_name, gpointer user_data, GError **error)
{
	appdata *data;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	if (g_str_equal(element_name, "domainList")) {
		/*Окончание списка*/
		data->xmlparameter = APPLICATION_XML_PARAMETER_UNKNOWN;
	} else if (g_str_equal(element_name, "domain")) {
		if ((data->current_whois_name != NULL) && (data->current_whois_server != NULL)) {
			if (data->current_whois_level > 1) {
				/*Добавление информации о серевре whois для доменной зоны третьего уровня в хэш-таблицу
				  (в этом случае уровень вложенности должен быть больше 1)*/
				g_hash_table_insert(data->whois_servers, g_strdup(data->current_whois_name), g_strdup(data->current_whois_server));
			} else {
				/*Освобождение памяти, использованной для хранения информации о серевере whois для доменной
				  зоны второго уровня*/
				g_free(data->current_whois_name);
				data->current_whois_name = NULL;
				g_free(data->current_whois_server);
				data->current_whois_server = NULL;
			}
		}
		/*Уменьшение текущего уровня вложенности*/
		data->current_whois_level--;
		data->xmlparameter = APPLICATION_XML_PARAMETER_UNKNOWN;
	}
}

/*Функция разбора файла со списком серверов whois формата XML*/
static gboolean application_parse_servers_xml(appdata *data)
{
	GMarkupParser parser;
	GMarkupParseContext *context;
	GBytes *xmlbytes;
	const gchar *contents;
	gsize length;
	GError *error;
		
	g_assert(data != NULL);
	
	/*Инициализация переменных для временного хранения данных*/
	data->current_whois_level = 0;
	data->current_whois_name = NULL;
	data->current_whois_server = NULL;
	
	/*Создание хэш-таблицы для хранения информации о сереверах whois и соотвествующих доменных зонах*/
	data->whois_servers = g_hash_table_new_full(g_str_hash, g_str_equal, application_free_domain_name, application_free_server_name);
	
	/*Загрузка данных файла со списком серверов whois из соотвествующего ресурса*/
	error = NULL;
	xmlbytes = g_resources_lookup_data(APPLICATION_SERVER_LIST_RESOURCE, G_RESOURCE_LOOKUP_FLAGS_NONE, &error);
	
	if (xmlbytes != NULL) {
		/*Установка указателя на функцию начала разбора элемента*/
		parser.start_element = application_xml_get_start_element;
		/*Установка указателя на функцию окончания разбора элемента*/
		parser.end_element = application_xml_get_end_element;
		/*Установка указателя на функцию разбора текста*/ 
		parser.text = application_xml_get_value;
		/*Реализация функции пропуска текста не используется*/
		parser.passthrough = NULL;
		/*Реализация функции обработки сообщения об ошибке не используется*/
		parser.error = NULL;
		/*Создание объекта контекста разбора документа формата XML*/
		context = g_markup_parse_context_new(&parser, 0, data, NULL);
		/*Получение побайтового доступа к данным документа*/
		contents = g_bytes_get_data(xmlbytes, &length);
		/*Непосредственный разбор документа*/
		g_markup_parse_context_parse(context, contents, (gssize)length, &error);
		if (error != NULL) {
			g_warning("Ошибка при обработке файла формата XML: %s\n", error->message);
			g_error_free(error);
		}
		/*Уничтожение объекта контекста разбора документа формата XML*/
		g_markup_parse_context_free(context);
		/*Освобождение памяти, использованной для хранения данных файла*/
		g_bytes_unref(xmlbytes);
	} else if (error != NULL) {
		g_warning("Ошибка при чтении файла формата XML: %s\n", error->message);
		g_error_free(error);
		return FALSE;			
	}
	
	return TRUE;
}

/*Функция изменения состояния графического интерфейса и вывода соответствующего сообщения
  в строке состояния*/
static void application_process_ui_change(appdata *data, guint processid, GError *error)
{
	gchar *statusmsg;
	
	g_assert(data != NULL);
	
	switch (processid) {
		case APPLICATION_PROCESS_START:
			/*Начало процесса получения инфорамции о домене*/
			statusmsg = g_strdup_printf("Получение информации о домене %s", gtk_entry_get_text(GTK_ENTRY(data->domain_entry)));
			gtk_widget_set_sensitive(data->search_button, FALSE);
			gtk_widget_set_sensitive(data->domain_entry, FALSE);
			gtk_widget_show(data->spinner);
			gtk_spinner_start(GTK_SPINNER(data->spinner));
			break;
		case APPLICATION_PROCESS_END:
			/*Окончание процесса получения информации о домене*/
			statusmsg = g_strdup_printf("Информация о домене %s получена", gtk_entry_get_text(GTK_ENTRY(data->domain_entry)));
			gtk_widget_set_sensitive(data->search_button, TRUE);
			gtk_widget_set_sensitive(data->domain_entry, TRUE);
			gtk_spinner_stop(GTK_SPINNER(data->spinner));
			gtk_widget_hide(data->spinner);
			break;
		case APPLICATION_PROCESS_ERROR:
			/*Ошибка в процессе получения информации о домене*/
			if (error != NULL) {
				statusmsg = g_strdup_printf("Ошибка: %s", error->message);
				g_error_free(error);
			} else {
				statusmsg = g_strdup_printf("Неизвестная ошибка");
			}
			gtk_widget_set_sensitive(data->search_button, TRUE);
			gtk_widget_set_sensitive(data->domain_entry, TRUE);
			gtk_spinner_stop(GTK_SPINNER(data->spinner));
			gtk_widget_hide(data->spinner);
			break;
		default:
			/*Неизвестное действие*/
			statusmsg = g_strdup_printf("Неизвестное действие");
			break;
	}
	
	/*Удаление предыдущей строки*/
	if (data->statusmsgid != 0) {
		gtk_statusbar_remove(GTK_STATUSBAR(data->statusbar), data->statusconid, data->statusmsgid);
	}
	/*Вывод строки*/
	data->statusmsgid = gtk_statusbar_push(GTK_STATUSBAR(data->statusbar), data->statusconid, statusmsg);
	/*Освобождение памяти*/
	g_free(statusmsg);
}

/*Функция обратного вызова для получения данных от серевера*/
static gboolean application_read_callback(GSocket *socket, GIOCondition cond, gpointer user_data)
{
	appdata *data;
	GError *error;
	gssize len;
	gchar buf[1024];
	GtkTextBuffer *textbuffer;
	GtkTextIter textiter;
	
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	if (cond == G_IO_IN) {
		/*Проверка выполнения установленного условия доступности данных (G_IO_IN)*/
		textbuffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(data->whois_textview));
		/*Получение текстового буфера виджета поля редактирования текста*/
		if (textbuffer != NULL) {
			/*Получение конечного итератора текстового буфера*/
			gtk_text_buffer_get_end_iter(textbuffer, &textiter);
			do {
				/*Получение всех передавнных сервером данных в цикле*/
				error = NULL;
				len = g_socket_receive(socket, buf, sizeof(buf), NULL, &error);
				if (len > 0) {
					/*Добавление полученного текста в текстовый буфер*/
					gtk_text_buffer_insert(textbuffer, &textiter, buf, len);
				} else if (len < 0) {
					/*Обработка ошибки, возникшей в процессе получения данных*/
					application_process_ui_change(data, APPLICATION_PROCESS_ERROR, error);
					g_object_unref(socket);
					return FALSE;
				}
			} while (len > 0);
		}
	} else {
		/*Обработка стандартных условий, соотвествующих ошибкам (G_IO_ERR, G_IO_HUP) */
		application_process_ui_change(data, APPLICATION_PROCESS_ERROR, NULL);
		g_object_unref(socket);
		return FALSE;
	}
	
	/*Завершение приема данных*/
	application_process_ui_change(data, APPLICATION_PROCESS_END, NULL);
	g_object_unref(socket);
	return FALSE;
}

/*Функция обратного вызова для обработки результатов разрешения доменных имен*/
static void application_resolver_lookup_callback(GObject *source_object, GAsyncResult *res, gpointer user_data)
{
	appdata *data;
	GError *error = NULL;
	GList *addresses;
	GInetAddress *inetaddress;
	GSocketAddress *socketaddress;
	GSocket *socket;
	GSource *source;
	gchar *request;
	gssize len;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Получение списка IP-адресов, соотвествующих заданному доменному имени*/
	addresses = g_resolver_lookup_by_name_finish(G_RESOLVER(source_object), res, &error);
	
	if (addresses != NULL) {
		/*Извлечение первого IP-адреса из списка (индекс первого объекта списка равен 0)*/
		inetaddress = G_INET_ADDRESS(g_list_nth_data(addresses, 0));
		/*Преобразование объекта IP-адреса в объект адреса сетевого сокета с указанием номера порта*/
		socketaddress = g_inet_socket_address_new(inetaddress, APPLICATION_WHOIS_PORT);
		/*Создание объекта сетевого сокета для работы с протоколом TCP*/
		socket = g_socket_new(g_inet_address_get_family(inetaddress), G_SOCKET_TYPE_STREAM, G_SOCKET_PROTOCOL_DEFAULT, &error);
		if (socket != NULL) {
			/*Установка ограничения по времени для операций с сетевым сокетом (равного 5 секундам)*/
			g_socket_set_timeout(socket, APPLICATION_SOCKET_TIMEOUT);
			if (g_socket_connect(socket, socketaddress, NULL, &error)) {
				/*Соединение сетевого сокета с сервером с заданным адресом*/
				request = g_strdup_printf("%s\r\n", gtk_entry_get_text(GTK_ENTRY(data->domain_entry)));
				/*Отправка запроса с именем домена*/
				len = g_socket_send(socket, (const gchar *)request, strlen(request), NULL, &error);
				/*Уничтожение объекта адреса сетевого сокета*/
				g_object_unref(socketaddress);
				g_free(request);
				if (len != -1) {
					/*Создание нового источника событий на основе сетевого сокета с указанием условия доступности данных для получения*/
					source = g_socket_create_source(socket, G_IO_IN, NULL);
					/*Установка функции обратного вызова, которая будет вызываться при выполнении заданного условия*/
					g_source_set_callback(source, (GSourceFunc)application_read_callback, data, NULL);
					/*Соединение созданного источника событий с главным циклом обработки событий приложения*/
					g_source_attach(source, NULL);
					/*Удаление ссылки на созданный источник событий для его автоматического уничтожения*/
					g_source_unref(source);
				} else {
					application_process_ui_change(data, APPLICATION_PROCESS_ERROR, error);
				}
			} else {
				application_process_ui_change(data, APPLICATION_PROCESS_ERROR, error);
			}
		} else {
			application_process_ui_change(data, APPLICATION_PROCESS_ERROR, error);
		}
		/*Уничтожение полученного списка IP-адресов*/
		g_resolver_free_addresses(addresses);
	} else {
		application_process_ui_change(data, APPLICATION_PROCESS_ERROR, error);
	}
}

/*Функция запуска процесса получения информации от сервера whois*/
static void application_start_whois_lookup(appdata *data)
{
	GtkTextBuffer *textbuffer;
	GtkTextIter startiter, enditer;
	GResolver *resolver;
	gchar *domain, *whois_server;
		
	g_assert(data != NULL);
	
	/*Получение текстового буфера виджета поля редактирования текста*/
	textbuffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(data->whois_textview));
	
	if (textbuffer != NULL) {
		/*Очистка текстового буфера*/
		gtk_text_buffer_get_bounds(textbuffer, &startiter, &enditer);
		gtk_text_buffer_delete(textbuffer, &startiter, &enditer);
	}
	
	/*Изменение состояния графического интерфейса*/
	application_process_ui_change(data, APPLICATION_PROCESS_START, NULL);
	
	/*Получение полного доменного имени*/
	domain = application_get_domain_name((gchar *)gtk_entry_get_text(GTK_ENTRY(data->domain_entry)));
	
	if (domain != NULL) {
		/*Получение имени сервера whois, предоставляющего информацию о заданном домене*/
		whois_server = g_hash_table_lookup(data->whois_servers, domain);
		if (whois_server != NULL) {
			/*Создание объекта разрешения доменных имен и разрешение доменного имени сервера whois в асинхронном режиме*/
			resolver = g_resolver_get_default();
			g_resolver_lookup_by_name_async(resolver, whois_server, NULL, application_resolver_lookup_callback, data);
		} else {
			application_process_ui_change(data, APPLICATION_PROCESS_ERROR, NULL);
		}
	} else {
		application_process_ui_change(data, APPLICATION_PROCESS_ERROR, NULL);
	}
}

/*Функция, вызываемая при нажатии кнопки Whois*/
static void application_search_button_clicked(GtkButton *button, gpointer user_data)
{
	appdata *data;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	application_start_whois_lookup(data);
}

/*Функция, вызывемая при нажатии клавиши Enter в поле ввода имени домена*/
static void application_domain_entry_activated(GtkEntry *entry, gpointer user_data)
{
	appdata *data;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	application_start_whois_lookup(data);
}

/*Обработчик сигнала изменения имени домена*/
static void application_domain_entry_changed(GtkEditable *editable, gpointer user_data)
{
	appdata *data;
	gchar *domain;
	
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Получение имени доменной зоны для заданного имени домена*/
	domain = application_get_domain_name(gtk_editable_get_chars(editable, 0, -1));
	
	if (domain != NULL) {
		/*Активация или деактивация механизма получения информации о домене в зависимости
		  от наличия в хэш-таблице сервера для заданной доменной зоны*/
		if (g_hash_table_contains(data->whois_servers, domain)) {
			if (!data->domain_entry_activation) {
				/*Активация механизма получения информации о домене*/
				gtk_widget_set_sensitive(data->search_button, TRUE);
				g_signal_handler_unblock(data->domain_entry, data->domain_entry_activate_signal);
				data->domain_entry_activation = TRUE;
			}
		} else {
			if (data->domain_entry_activation) {
				/*Деактивация механизма получения информации о домене*/
				gtk_widget_set_sensitive(data->search_button, FALSE);
				g_signal_handler_block(data->domain_entry, data->domain_entry_activate_signal);
				data->domain_entry_activation = FALSE;
			}
		}
	}
}

/*Функция для создания главного окна приложения*/
static void application_activate(GtkApplication *application, gpointer user_data)
{
	appdata *data;
	GtkWidget *vbox, *hbox;
	GtkWidget *scrollwin;
	GdkDisplay *display;
	GdkScreen *screen;
	GtkCssProvider *provider;
		
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Разбор файла со списком серверов whois формата XML*/
	application_parse_servers_xml(data);
	
	/*Создание главного окна*/
	data->mainwin = gtk_application_window_new(application);
	gtk_window_set_title(GTK_WINDOW(data->mainwin), "Пример использования класса GSocket (протокол TCP) и ресурсов приложения");
	gtk_widget_set_size_request(data->mainwin, 600, 400);
	/*Создание контейнера для хранения элементов главного окна*/
	vbox = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
	/*Создание контейнера для элементов управления*/
	hbox = gtk_box_new(GTK_ORIENTATION_HORIZONTAL, 0);
	/*Создание виджета для ввода имени домена*/
	data->domain_entry = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(data->domain_entry), "kernel.org");
	/*Флаг активации механизма получения информации о домене*/
	data->domain_entry_activation = TRUE;
	/*Установка обработчика сигнала активации поля ввода имени домена и сохранение его идентификатора*/
	data->domain_entry_activate_signal = g_signal_connect(data->domain_entry, "activate", G_CALLBACK(application_domain_entry_activated), data);
	/*Установка обработчика сигнала изменения текста в поле ввода имени домена*/
	g_signal_connect(data->domain_entry, "changed", G_CALLBACK(application_domain_entry_changed), data);
	gtk_box_pack_start(GTK_BOX(hbox), data->domain_entry, TRUE, TRUE, 5);
	/*Создание виджета кнопки для получения информации*/
	data->search_button = gtk_button_new_with_label("Whois");
	/*Устанока сигнала нажатия на кнопку для получения информации*/
	g_signal_connect(data->search_button, "clicked", G_CALLBACK(application_search_button_clicked), data);
	gtk_box_pack_start(GTK_BOX(hbox), data->search_button, FALSE, TRUE, 5);
	/*Добавление контейнера с элементами управления в основной контейнер*/
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, TRUE, 5);
	/*Создание окна для прокрутки содержимого виджета поля для редактирования текста*/
	scrollwin = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrollwin), GTK_SHADOW_IN);
	/*Создание виджета поля для редактирования текста*/
	data->whois_textview = gtk_text_view_new();
	gtk_container_add(GTK_CONTAINER(scrollwin), data->whois_textview);
	gtk_text_view_set_editable(GTK_TEXT_VIEW(data->whois_textview), FALSE);
	gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(data->whois_textview), GTK_WRAP_WORD);
	gtk_box_pack_start(GTK_BOX(vbox), scrollwin, TRUE, TRUE, 0);
	/*Создание виджета строки состояния*/
	data->statusbar = gtk_statusbar_new();
	data->statusmsgid = 0;
	data->statusconid = gtk_statusbar_get_context_id(GTK_STATUSBAR(data->statusbar), "GWhois");
	/*Создание виджета индикатора выполнения задачи*/
	data->spinner = gtk_spinner_new();
	/*Добавление виджета индикатора выполнения задачи в строку состояния*/
	gtk_box_pack_end(GTK_BOX(data->statusbar), data->spinner, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), data->statusbar, FALSE, TRUE, 0);
	gtk_container_add(GTK_CONTAINER(data->mainwin), GTK_WIDGET(vbox));
	/*Показ главного окна приложения*/
	gtk_application_add_window(application, GTK_WINDOW(data->mainwin));
	gtk_widget_show_all(data->mainwin);
	/*Скрытие виджета индикатора выполнения задачи*/
	gtk_widget_hide(data->spinner);
	
	/*Получение объекта текущего дисплея*/
	display = gdk_display_get_default();
	if (display != NULL) {
		/*Получение объекта экрана для текущего дисплея*/
		screen = gdk_display_get_default_screen(display);
		/*Создание объекта системы разбора каскадных таблиц стилей и его активация*/
		provider = gtk_css_provider_new();
		gtk_style_context_add_provider_for_screen(screen, GTK_STYLE_PROVIDER(provider), GTK_STYLE_PROVIDER_PRIORITY_USER);
		/*Загрузка каскадной таблицы стилей из ресурса для различных версий GTK*/
		#if GTK_CHECK_VERSION(3, 16, 0)
			/*В версии GTK 3.16 появился отдельный метод для загрузки каскадных таблиц стилей из ресурсов*/
			gtk_css_provider_load_from_resource(GTK_CSS_PROVIDER(provider), APPLICATION_CSS_RESOURCE);
		#else
			GBytes *cssbytes;
			const gchar *cssdata;
			gsize csslength;
			GError *error;
			/*В более ранних версиях GTK загрузка каскадных таблиц стилей из ресурсов должна осуществляться вручную*/
			error = NULL;
			/*Получение данных ресурса*/
			cssbytes = g_resources_lookup_data(APPLICATION_CSS_RESOURCE, G_RESOURCE_LOOKUP_FLAGS_NONE, &error);
			if (cssbytes != NULL) {
				/*Получение побайтового доступа к данным*/
				cssdata = g_bytes_get_data(cssbytes, &csslength);
				/*Загрузка данных каскадной таблицы стилей*/
				if (!gtk_css_provider_load_from_data(GTK_CSS_PROVIDER(provider), cssdata, csslength, &error)) {
					g_warning("Ошибка загрузки данных каскадной таблицы стилей: %s\n", error->message);
					g_error_free(error);
				}
				/*Уничтожение объекта данных (обратите внимание на то, что в данном случае не используется метод g_object_unref())*/
				g_bytes_unref(cssbytes);
			} else {
				g_warning("Ошибка загрузки данных ресурса: %s\n", error->message);
				g_error_free(error);
			}
		#endif
		g_object_unref(provider);
	}
}

/*Функция, вызываемая перед завершением работы приложения*/
static void application_shutdown(GtkApplication *application, gpointer user_data)
{
	appdata *data;
	
	g_assert(user_data != NULL);
	
	data = (appdata *)user_data;
	
	/*Уничтожение хэш-таблицы с именами серверов whois*/
	g_hash_table_destroy(data->whois_servers);
}

int main(int argc, char **argv)
{
	appdata *data;
	gint res;
	
	/*Создание структуры для хранения данных приложения*/
	data = g_new0(appdata, 1);
	/*Создание объекта приложения*/
	data->app = gtk_application_new("org.gtk.GWhois", G_APPLICATION_FLAGS_NONE);
	/*Установка обработчиков основных сигналов объекта приложения*/
	g_signal_connect(data->app, "activate", G_CALLBACK(application_activate), data);
	g_signal_connect(data->app, "shutdown", G_CALLBACK(application_shutdown), data);
	/*Запуск основного цикла обработки событий приложения*/
	res = g_application_run(G_APPLICATION(data->app), 0, NULL);
	/*Освобождение памяти перед завершением работы приложения*/
	g_free(data);
	
	return res;
}
